// (c) Kyros Kutulakos 2/06/01
//
// Adapted for use in CSC290C course
//			list of revisions:
//				+ implemented a set() method that changes the image being
//				  displayed & eliminates the need for the dispImage global variable
//				+ illustrates how textured and non-textured objects can be displayed
//				  simultaneously using the Texture.activate()/deactivate() methods
//				+ added an event handler for getting mouse-1 input to select a rectangular
//				  region in the image
//				+ allows the window to be resized by intercepting the window-resize event
//				  with the resize() method
//				+ illustrates how to use the function make_current() to direct all OpenGL
//				  commands to the current OpenGL panel (useful if you have multiple panels 
//				  showing at once)
//			    + changed the glOrtho parameters and the coordinates of the texture-mapped
//				  polygon. this is necessary if one wants point (x,y) drawn on the screen
//			      using OpenGL to match point (x,y) in Vista coordinates: this is because
//				  Vista coordinates have the (0,0) point at the top-left whereas OpenGL
//				  coordinates have it at the bottom left. 
//

// This file gives an example of how an image can be displaed in an OpenGL panel
// which is part of an fltk window

// The idea behind the image display code below is to display an image by 
// drawing a rectangle on the screen whose dimensions are equal to the initial size 
// of the OpenGL panel. Rather than giving the rectangle a single color, however, we
// texture-map the Vista image to be displayed in that rectangle. 
//
// Doing the image display in this way will be useful for a number of assignments
// later on, because it allows us to shrink, distort, enlarge, or warp an image
// by simply changing the size and position of the rectancle's 4 vertices.

// To better understand the code in this file, you may want to refer to the
// Texture Mapping section of the OpenGL programming guide 


#include <math.h>
#include <iostream>
#include <Fl/Fl_File_Chooser.H>
#include <Fl/Fl_Color_Chooser.H>
#include "imdraw.h"
#include "Texture.h"
#include "glutils.h"
#include "Line.h"

// Shape type if select was chosen
#define SELECT 2

ImDraw::ImDraw(int x,int y,int w,int h,const char *l)
            : Fl_Gl_Window(x,y,w,h,l)
{
  /* w and h are the width and height of the OpenGL canvas */
  width = w;
  height = h;
  im_width = width;
  im_height = height;
  aspect_ratio = 1;
  
  /* Initially, the allocated texture memory contains no data */
  imTexture = NULL;

  // Reset the left mouse button state.
  mousedown = 0;

  // Set the current shape type to LINE.
  set_line();

  // Initialize the selected line flag to be false.
  is_selected = 0;
}

void ImDraw::mypartner(ImDraw *partner) {
  this->partner = partner;
}

// Sets the image being displayed in the window
void ImDraw::set(VImage	new_im)
{
	if ((VImageNRows(new_im) < 1) || (VImageNColumns(new_im) < 1)) {
		return;
	}

	// Direct all further OpenGL commands to this OpenGL panel
	// This is NECESSARY if there are more than one OpenGL panels on the screen!
	window()->make_current();
	if (imTexture == NULL) {
	  /* These commands are executed only once, when the variable containing
         the image to be displayed points to a valid Vista image.
        The command initializes the OpenGL texturing operations
      */

	
	  /* Allocate space in texture memory for a single texture that has 
         a maximum size of 1024x1024
  
         Note that texture dimensions have to be powers of 2 and that
         the dimensions of any Vista image to be used as texture must
         have dimensions smaller than the texture dimensions. 

         You will need to call this routine with different parameters if you are
         to display Vista images larger than 1024x1024 (this is probably large 
	 enough and you might want to use smaller images to conserve memory & disk
	 space
      */
	  texture_set_globals(1024, 1024, 3);
	  imTexture = new Texture(new_im);
	} else
	  imTexture->refresh(new_im);

	aspect_ratio = VImageNColumns(new_im)*1.0/VImageNRows(new_im);
	im_height = VImageNRows(new_im);
	im_width = VImageNColumns(new_im);

    // Redraw the entire window to show the new image
	Fl_Window::redraw();
}
	

/* The following routine is the main drawing routine */ 

/* This is where you must add code to do whatever drawing operations the user
   requests
   
   The routine below currently displays a Vista image in the GL drawing area,
   giving you an example of how images can be drawn onto the GL screen

   Note that every time draw is called, it must draw
   ALL the images/lines/etc that the user has drawn already 
*/

void ImDraw::draw() {
  double tex_width, tex_height;
  double w, h, mag;

  window()->make_current();
  if (!valid()) {
    /* These commands just tell OpenGL what the size of the canvas is */
    glLoadIdentity(); glViewport(0,0,width,height);
	// This definition was changed from the previous version so that both the 
	// OpenGL and Vista coordinates draw the (0,0) point at the top-left of the window
    glOrtho(0, width, height, 0, -1.0, 1.0);
  }

  /* Use black to clear canvas */
  glClearColor(0.0, 0.0, 0.0, 1.0);
  /* Clear the gl canvas */
  glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);

  //
  //
  // -- the code fragment below is for displaying a vista image in an OpenGL --
  //
  //

  //
  // Display image by texture-mapping it onto a polygon whose size is equal to the
  // size of the GL panel
  //

  
  /* texture-map polygons with the current texture */
  if (imTexture != NULL) {
    /* Store the current contents of dispImage in texture memory */
    //imTexture->refresh(dispImage);
    /* Record the texture coordinates of the top-right corner of the image */
    imTexture->get_bounds(tex_width, tex_height);
    /* Enable GL texture mapping */
    imTexture->activate();
  }
  

  // define the width and height of the canvas area that will be occupied by the 
  // displayed image, in order to preserve the image's aspect ratio (which may be
  // different from the aspect ratio of the canvas window)
  h = (height < width/aspect_ratio)?height:(width/aspect_ratio);
  w = h*aspect_ratio;
  mag = w/im_width;

  // The following statements are what actually display the image in the panel.
 
  glColor3d(0.0, 0.0, 0.0);
  glBegin(GL_POLYGON);
  {
    glTexCoord2d(0,0);
    glVertex2d(0,h);
    
    glTexCoord2d(0,tex_height);
    glVertex2d(0,0);
    
    glTexCoord2d(tex_width, tex_height);
    glVertex2d(w, 0);
    
    glTexCoord2d(tex_width,0);
    glVertex2d(w, h);
  }
  glEnd();
  glFlush();
  
  // Disable texture mapping for the OpenGL commands that follow
  if (imTexture != NULL) {
     imTexture->deactivate();
  }

  // Draw all the previously completed shapes.
  for (list<Shape *>::iterator i = shapes.begin(); i!=shapes.end(); i++) {
    (*i)->draw(mag);
  }
  
  // Check if another shape is in process of being drawn.
  if (mousedown && curshape != NULL) {
    curshape->draw(mag);
  }
}

// Routine that intercepts window resize operations performed by the user
void ImDraw::resize(int x, int y, int w, int h)
{
  // the window has been resized so we need to update the width & height 
  // information 
  width = w;
  height = h;
  
  // Need to call this method ensure that the GL panel "knows" that it has
  // been resized as well...
  Fl_Gl_Window::resize(x, y, w, h);
}

// define an event handler that intercepts Mouse-1 events 
int ImDraw::handle(int event) {
  int mousex, mousey; 
  int intercepted=0;
  double h = (height < width/aspect_ratio)?height:(width/aspect_ratio);
  double w = h*aspect_ratio;
  double mag = w/im_width;
  // A flag that checks if a line was selected.
  int no_selection = 1;

  switch (event) {
  case FL_PUSH:
	// these commands get the position of the mouse when the mouse event
	// occured
    mousex = Fl::event_x();
    mousey = Fl::event_y();
    intercepted = 1;
    if (Fl::event_button() == 1) {
      // Create a new shape if left mouse button is pushed.
      switch (shapetype) {
        case LINE:
	  curshape = new Line();
	  x1 = x2 =  mousex;
	  y1 = y2 = mousey;
	  break;
        case SELECT:
	  // Start draging the line or its edges if a line was selected.
	  if (is_selected) {
	    // Check if the edge point or the inside of the line was selected.
	    switch (what_selected) {
	    case 1:
	      x2 = int(((Line *)*selected)->get_x2() * mag);
	      y2 = int(((Line *)*selected)->get_y2() * mag);
	      x1 = mousex;
	      y1 = mousey;
	      break;
	    case 2:
	      x1 = int(((Line *)*selected)->get_x1() * mag);
	      y1 = int(((Line *)*selected)->get_y1() * mag);
	      x2 = mousex;
	      y2 = mousey;
	      break;
	    }
	  }
      }
      mousedown = 1;
    } else if (is_selected && shapetype == SELECT && Fl::event_button() == 2) {
      // Delete the selected line if the middle mouse button is pressed on it.
      int index = 0;
      for (list<Shape *>::iterator i = shapes.begin(); i!=selected; i++) {
	index++;
      }

      partner->deleteshape(index);
      partner->redraw();
      delete (*selected);
      shapes.erase(selected);
      is_selected = 0;
    }
    break;
  case FL_RELEASE:
    // Finish updating the new shape if left mouse button is released.
    // Insert the finished shape into the list of shapes.
    mousex = Fl::event_x();
    mousey = Fl::event_y();
    intercepted = 1;
    if (Fl::event_button() == 1 && shapetype != SELECT) {
      shapes.push_back(curshape);
      if (curshape->get_type() == LINE) {
	Shape *copyshape = new Line((Line *)curshape);
	partner->addshape(copyshape);
	partner->redraw();
      }
      curshape = NULL;
      mousedown = 0;
    }
    break;
  case FL_DRAG:
    // Update the current mouse coordinates if the mouse is draged
    // in order to later update the current shape.
    mousex = Fl::event_x();
    mousey = Fl::event_y();
    intercepted = 1;
    if (mousedown) {
      // An edge point is being draged
      if (shapetype == SELECT && is_selected) {
	switch (what_selected) {
	case 1:
	  x1 = mousex;
	  y1 = mousey;
	  break;
	case 2:
	  x2 = mousex;
	  y2 = mousey;
	  break;
	case 3:
	  x1 = mousex + delta_x1*mag;
	  x2 = mousex + delta_x2*mag;
	  y1 = mousey + delta_y1*mag;
	  y2 = mousey + delta_y2*mag;
	}
	((Line *)*selected)->setcoords(x1/mag,y1/mag,x2/mag,y2/mag);
      }
      if (shapetype == LINE) {
	x2 = mousex;
	y2 = mousey;
	((Line *)curshape)->setcoords(x1/mag,y1/mag,x2/mag,y2/mag);
      }
    }
    
    break;
  case FL_MOVE:
    mousex = Fl::event_x();
    mousey = Fl::event_y();
    intercepted = 1;

    // Check if a line is selected
    if (shapetype == SELECT) {
      for (list<Shape *>::iterator i = shapes.begin(); i!=shapes.end(); i++) {
	if ((*i)->get_type() == LINE) {
	  switch (((Line *)(*i))->mouse_on(mousex*1.0/mag,mousey*1.0/mag,int(10/mag))) {
	    // The first endpoint was selected
	  case 1:
	    if (!no_selection) {
	      ((Line *)(*selected))->set_handle_status(0);
	    }
	    ((Line *)(*i))->set_handle_status(1);
	    selected = i;
	    what_selected = 1;
	    no_selection = 0;
	    break;
	    // The second endpoint was selected
	  case 2:
	    if (!no_selection) {
	      ((Line *)(*selected))->set_handle_status(0);
	    }
	    ((Line *)(*i))->set_handle_status(1);
	    selected = i;
	    what_selected = 2;
	    no_selection = 0;
	    break;
	    // The inside of the line was selected
	  case 3:
	    if (!no_selection) {
	      ((Line *)(*selected))->set_handle_status(0);
	    }
	    ((Line *)(*i))->set_handle_status(1);
	    selected = i;
	    what_selected = 3;
	    delta_x1 = ((Line *)(*i))->get_x1() - mousex/mag;
	    delta_x2 = ((Line *)(*i))->get_x2() - mousex/mag;
	    delta_y1 = ((Line *)(*i))->get_y1() - mousey/mag;
	    delta_y2 = ((Line *)(*i))->get_y2() - mousey/mag;
	    no_selection = 0;
	    break;
	  case 0:
	    ((Line *)(*i))->set_handle_status(0);
	    break;
	  }
	}
      }
    }
    is_selected = !no_selection;
    break;
  default:
    ;
  }
  if (!intercepted)
    // delegate all other events to the regular Fl_Gl_Window handler
    return(Fl_Gl_Window::handle(event));
  else {
    redraw();
    return 1;
  }
}

void ImDraw::set_line() {
  shapetype = LINE;
}

void ImDraw::set_select() {
  shapetype = SELECT;
}

char* ImDraw::Load() {
  char* filename = NULL;

  // Choose a vista file from a file chooser dialog.
  filename = fl_file_chooser("Load file ... ","*.v","alex.v");
  if (filename != NULL) {
    FILE* imfile = fopen(filename,"rb");
    
    // Read the Vista Image
    if (VReadImages(imfile,&attrs, &im) != 0) {
      set(im[0]);
      fclose(imfile);
    }
  }
  return filename;
}

char* ImDraw::Save() {
  char* filename;
  fstream f;
  int type;
  int size = shapes.size();

  // Choose a file from a file chooser dialog.
  filename = fl_file_chooser("Save to file ... ","*.ld","mydrawing.ld");

  // Check if CANCEL wasn't pressed in the dialog.
  if (filename != NULL) {
    f.open(filename,ios::out | ios::binary);
    f.write((char *)&size,sizeof(int));

    // Save every shape in the shape list to a file.
    for (list<Shape *>::iterator i = shapes.begin(); i!=shapes.end(); i++) {

      // Write the shape type first.
      type = (*i)->get_type();
      f.write((char *)&type, sizeof(int));

      // Write the shape information after its type was written.
      switch (type) {
      case LINE:
	(*i)->write(f);
	break;
      case CIRCLE:
	(*i)->write(f);
	break;
      }
    }
    f.close();
  }

  return filename;
}  

void ImDraw::delete_shapes() {
  for (list<Shape *>::iterator i = shapes.begin(); i!=shapes.end(); i++) {
    delete (*i);
  }
}

void ImDraw::New() {
  delete_shapes();
  shapes.clear();
  redraw();
}

void ImDraw::addshape(Shape* shape) {
  shapes.push_back(shape);
}

void ImDraw::deleteshape(int index) {
  int count = 0;
  list<Shape *>::iterator i;
  for (i = shapes.begin(); count < index; i++) {
    count++;
  }

  if (i == selected) {
    is_selected = 0;
  }
  delete (*i);
  shapes.erase(i);
}

VImage* ImDraw::getimage() {
  return im;
}

list<Shape *> *ImDraw::getshapes() {
  return &shapes;
}
